"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class App(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    FLAGS_FIELD_NUMBER: builtins.int
    HASH_FIELD_NUMBER: builtins.int
    HASHCODEDATA_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    flags: builtins.int
    hash: builtins.bytes
    hashCodeData: builtins.bytes
    name: typing.Text
    def __init__(self,
        *,
        flags: builtins.int = ...,
        hash: builtins.bytes = ...,
        hashCodeData: builtins.bytes = ...,
        name: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["flags",b"flags","hash",b"hash","hashCodeData",b"hashCodeData","name",b"name"]) -> None: ...
global___App = App

class AppList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    LIST_FIELD_NUMBER: builtins.int
    @property
    def list(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___App]: ...
    def __init__(self,
        *,
        list: typing.Optional[typing.Iterable[global___App]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["list",b"list"]) -> None: ...
global___AppList = AppList



import logging
import struct
from typing import Union

from construct import (
    Bytes,
    Const,
    GreedyRange,
    Hex,
    Int8ub,
    Int32ub,
    PascalString,
    Rebuild,
    Struct,
    len_,
    this,
)
from intelhex import IntelHex

from ledgerwallet.crypto.ecc import PrivateKey
from ledgerwallet.crypto.scp import SCP, FakeSCP
from ledgerwallet.hsmscript import HsmScript
from ledgerwallet.hsmserver import HsmServer
from ledgerwallet.ledgerserver import LedgerServer
from ledgerwallet.manifest import AppManifest
from ledgerwallet.proto.listApps_pb2 import AppList
from ledgerwallet.simpleserver import SimpleServer
from ledgerwallet.transport import FileDevice, enumerate_devices
from ledgerwallet.utils import LedgerIns, LedgerSecureIns, VersionInfo, serialize

LOAD_SEGMENT_CHUNK_HEADER_LENGTH = 3
MIN_PADDING_LENGTH = 1
SCP_MAC_LENGTH = 0xE

MAX_CHUNK_SIZE = (
    0x10000  # Maximum size of a chunk that can be loaded at once during app install
)

LEDGER_HSM_URL = "https://hsmprod.hardwarewallet.com/hsm/process"
LEDGER_HSM_KEY = "perso_11"


ApduListAppsResponse = Struct(
    Const(b"\x01"),  # Version
    apps=GreedyRange(
        Struct(
            # Application
            # Prefixed by the size of the structure, size included.
            _size=Rebuild(Int8ub, 1 + 4 + 32 + 32 + len_(this.name)),
            flags=Hex(Int32ub),
            code_data_hash=Bytes(32),
            full_hash=Bytes(32),
            name=PascalString(Int8ub, "utf-8"),
        )
    ),
)


class AppInfo(object):
    def __init__(
        self,
        name: str,
        flags: int,
        code_data_hash: bytes,
        full_hash: bytes,
        data: bytes = None,
    ):
        self.name = name
        self.flags = flags
        self.code_data_hash = code_data_hash
        self.full_hash = full_hash
        self.data = data


class MemoryInfo:
    def __init__(
        self,
        system_size: int,
        applications_size: int,
        free_size: int,
        used_app_slots: int,
        num_app_slots: int,
    ):
        self.system_size = system_size
        self.applications_size = applications_size
        self.free_size = free_size
        self.used_app_slots = used_app_slots
        self.num_app_slots = num_app_slots


class CommException(Exception):
    def __init__(self, message, sw=0x6F00, data=None):
        self.message = message
        self.sw = sw
        self.data = data

    def __str__(self):
        buf = "Exception : " + self.message
        return buf


class NoLedgerDeviceException(Exception):
    pass


LOG = logging.getLogger("ledgerwallet")


class LedgerClient(object):
    def __init__(self, device=None, cla=0xE0, private_key=None):
        self.scp = None
        if device is None:
            devices = enumerate_devices()
            if len(devices) == 0:
                raise NoLedgerDeviceException("No Ledger device has been found.")
            device = devices[0]
        elif type(device) == FileDevice:
            self.scp = FakeSCP()

        self.device = device
        self.cla = cla
        self._target_id = None
        if private_key is None:
            self.private_key = PrivateKey()
        else:
            self.private_key = PrivateKey(private_key)
        self.device.open()

    def close(self):
        self.device.close()

    def raw_exchange(self, data: bytes) -> bytes:
        LOG.debug("=> " + data.hex())
        output_data = bytes(self.device.exchange(data))
        if len(output_data) > 0:
            LOG.debug("<= " + output_data.hex())
        return output_data

    def apdu_exchange(self, ins, data=b"", p1=0, p2=0):
        apdu = bytes([self.cla, ins, p1, p2])
        apdu += serialize(data)
        response = self.raw_exchange(apdu)

        status_word = int.from_bytes(response[-2:], "big")
        if status_word != 0x9000 and ((status_word >> 8) != 0x61):
            possible_cause = "Unknown reason"
            if status_word == 0x6982:
                possible_cause = (
                    "Have you uninstalled the existing CA with resetCustomCA first?"
                )
            elif status_word == 0x6985:
                possible_cause = "Condition of use not satisfied (denied by the user?)"
            elif status_word == 0x6A84 or status_word == 0x6A85:
                possible_cause = "Not enough space?"
            elif status_word == 0x6A83:
                possible_cause = (
                    "Maybe this app requires a library to be installed first?"
                )
            elif status_word == 0x6484:
                possible_cause = "Are you using the correct targetId?"
            elif status_word == 0x69D5:
                possible_cause = "Cannot create custom secure channels on this device"
            raise CommException(
                "Invalid status %04x (%s)" % (status_word, possible_cause),
                status_word,
                data,
            )

        return response[:-2]

    def apdu_secure_exchange(self, ins, data=b"", p1=0, p2=0):
        if self.scp is None:
            server = SimpleServer(self.private_key)
            secret = self.authenticate(server)
            self.scp = SCP(secret)

        data = self.apdu_exchange(
            LedgerIns.SECUINS, self.scp.wrap(bytes([ins]) + data), p1, p2
        )
        return self.scp.unwrap(data)

    def authenticate(self, server: LedgerServer):
        self.reset()
        if self.target_id & 0xF < 2:
            raise Exception("Target ID does not support SCP V2")

        # Exchange nonce
        server_nonce = server.get_nonce()
        data = self.apdu_exchange(LedgerIns.INITIALIZE_AUTHENTICATION, server_nonce)
        device_nonce = data[4:12]
        server.send_nonce(device_nonce)

        # Get server certificate chain
        server_chain = server.receive_certificate_chain()
        for i in range(len(server_chain)):
            if i == len(server_chain) - 1:
                self.apdu_exchange(
                    LedgerIns.VALIDATE_CERTIFICATE, server_chain[i], p1=0x80
                )
            else:
                self.apdu_exchange(LedgerIns.VALIDATE_CERTIFICATE, server_chain[i])

        # Walk the client chain
        client_chain = []
        for i in range(2):
            if i == 0:
                certificate = self.apdu_exchange(LedgerIns.GET_CERTIFICATE)
            else:
                certificate = self.apdu_exchange(LedgerIns.GET_CERTIFICATE, p1=0x80)
            if len(certificate) == 0:
                break
            client_chain.append(certificate)
        server.send_certificate_chain(client_chain)

        # Mutual authentication done, retrieve shared secret
        self.apdu_exchange(LedgerIns.MUTUAL_AUTHENTICATE)
        return server.get_shared_secret()

    def _load_chunk(self, hex_file: IntelHex, segment, offset: int):
        start_addr, end_addr = segment
        segment_load_address = start_addr - hex_file.minaddr()

        self.apdu_secure_exchange(
            LedgerSecureIns.SET_LOAD_OFFSET,
            struct.pack(">I", segment_load_address + offset),
        )

        load_size = min(end_addr - start_addr - offset, MAX_CHUNK_SIZE)
        # max_load_size = 0xf0 - LOAD_SEGMENT_CHUNK_HEADER_LENGTH - MIN_PADDING_LENGTH - SCP_MAC_LENGTH # noqa
        max_load_size = 0x80

        load_address = start_addr + offset
        chunk_offset = start_addr
        while load_size > 0:
            chunk_size = min(load_size, max_load_size)
            data = hex_file.gets(load_address, chunk_size)
            data = struct.pack(">H", chunk_offset - start_addr) + data

            self.apdu_secure_exchange(LedgerSecureIns.LOAD, data)
            load_address += chunk_size
            chunk_offset += chunk_size
            load_size -= chunk_size

    def _load_segment(self, hex_file: IntelHex, segment):
        start_addr, end_addr = segment
        # Load each segment by chunks of 64kB
        for offset in range(0, end_addr - start_addr, MAX_CHUNK_SIZE):
            self._load_chunk(hex_file, segment, offset)

    def install_app(self, app_manifest: AppManifest):
        version_info = self.get_version_info()
        device = str(version_info.target_id)

        app_manifest.assert_compatible_device(version_info.target_id)

        hex_file = IntelHex(app_manifest.get_binary(device))
        code_length = hex_file.maxaddr() - hex_file.minaddr() + 1
        data_length = app_manifest.data_size(device)

        code_length -= data_length
        assert code_length % 64 == 0  # code length must be aligned

        flags = app_manifest.get_application_flags(device)  # not handled yet

        params = app_manifest.serialize_parameters(device)
        main_address = hex_file.start_addr["EIP"] - hex_file.minaddr()

        level = app_manifest.get_api_level(device)
        if level:
            data = struct.pack(
                ">BIIIII",
                level,
                code_length,
                data_length,
                len(params),
                flags,
                main_address,
            )
        else:
            data = struct.pack(
                ">IIIII", code_length, data_length, len(params), flags, main_address
            )
        self.apdu_secure_exchange(LedgerSecureIns.CREATE_APP, data)

        hex_file.puts(hex_file.maxaddr() + 1, params)
        for segment in hex_file.segments():
            self._load_segment(hex_file, segment)
        self.apdu_secure_exchange(LedgerSecureIns.COMMIT)

    def delete_app(self, app: Union[str, bytes]):
        if isinstance(app, str):
            self.apdu_secure_exchange(
                LedgerSecureIns.DELETE_APP, serialize(app.encode())
            )
        elif isinstance(app, bytes) and len(app) == 32:
            self.apdu_secure_exchange(LedgerSecureIns.DELETE_APP_BY_HASH, app)
        else:
            raise TypeError("app parameter must be string or digest")

    def install_remote_app(
        self, app_path, key_path, url=LEDGER_HSM_URL, key=LEDGER_HSM_KEY
    ):
        script = HsmScript("distributeFirmware11", {"persoKey": key, "scpv2": "dummy"})
        server = HsmServer(script, url)
        self.authenticate(server)

        application_data = server.query(
            params={"firmware": app_path, "firmwareKey": key_path, "scpv2": "dummy"}
        )
        offset = 0
        while offset < len(application_data):
            apdu_len = application_data[offset + 4]
            self.raw_exchange(application_data[offset : offset + 5 + apdu_len])
            offset += 5 + apdu_len

    def delete_remote_app(
        self, app_path, key_path, url=LEDGER_HSM_URL, key=LEDGER_HSM_KEY
    ):
        return self.install_remote_app(app_path, key_path, url, key)

    def upgrade_firmware(
        self, firmware_name, firmware_key, url=LEDGER_HSM_URL, key=LEDGER_HSM_KEY
    ):
        script = HsmScript(
            "distributeFirmware11_scan", {"persoKey": key, "scpv2": "dummy"}
        )
        server = HsmServer(script, url)
        self.authenticate(server)

        response = b""
        while True:
            app_data = server.query(
                data=None if len(response) < 2 else b"\xff\xff" + response[:-2]
            )
            if len(app_data) == 0:
                break
            response = self.raw_exchange(app_data)

        application_data = server.query(
            params={"firmware": firmware_name, "firmwareKey": firmware_key}
        )
        offset = 0
        while offset < len(application_data):
            apdu_len = application_data[offset + 4]
            self.raw_exchange(application_data[offset : offset + 5 + apdu_len])
            offset += 5 + apdu_len

    def genuine_check(self, url=LEDGER_HSM_URL, key=LEDGER_HSM_KEY):
        script = HsmScript("checkGenuine", {"persoKey": key, "scpv2": "dummy"})
        server = HsmServer(script, url)
        self.authenticate(server)

        client_data = b""
        while True:
            application_data = server.query(client_data[:-2])  # No not send status word
            if len(application_data) < 5:
                break
            client_data = self.raw_exchange(application_data)
        # custom_ui = client_data[0]
        # custom_ca = client_data[1]
        return True

    def endorse(self, key_id: int, url=LEDGER_HSM_URL, key=LEDGER_HSM_KEY):
        script = HsmScript("signEndorsement", {"persoKey": key})
        server = HsmServer(script, url)
        self.authenticate(server)
        server.query()  # Commit agreement

        data = self.apdu_exchange(LedgerIns.ENDORSE_SET_START, p1=key_id)
        certificate = server.query(data, params={"endorsementKey": "attest_1"})

        # Commit endorsement certificate
        self.apdu_exchange(LedgerIns.ENDORSE_SET_COMMIT, certificate)
        return True

    def install_ca(self, name: str, public_key: bytes):
        data = serialize(name.encode()) + serialize(public_key)
        self.apdu_secure_exchange(LedgerSecureIns.SETUP_CUSTOM_CERTIFICATE, data)

    def delete_ca(self):
        return self.apdu_secure_exchange(LedgerSecureIns.RESET_CUSTOM_CERTIFICATE)

    def get_version_info(self):
        data = self.apdu_exchange(LedgerIns.GET_VERSION)
        version_info = VersionInfo.parse(data)
        self._target_id = version_info.target_id
        return version_info

    def get_version_info_secure(self):
        data = self.apdu_secure_exchange(LedgerSecureIns.GET_VERSION)
        version_info = VersionInfo.parse(data)
        self._target_id = version_info.target_id
        return version_info

    def validate_target_id(self, target_id: int):
        self.apdu_exchange(LedgerIns.VALIDATE_TARGET_ID, struct.pack(">I", target_id))

    def reset(self):
        return self.validate_target_id(self.target_id)

    def get_memory_info(self) -> MemoryInfo:
        response = self.apdu_secure_exchange(LedgerSecureIns.GET_MEMORY_INFORMATION)
        assert len(response) == 20

        return MemoryInfo(*struct.unpack(">IIIII", response))

    @property
    def target_id(self):
        if self._target_id is None:
            self.get_version_info()
        return self._target_id

    def list_apps_remote(self, url=LEDGER_HSM_URL, key=LEDGER_HSM_KEY):
        script = HsmScript("listApps", {"persoKey": key, "scpv2": "dummy"})
        server = HsmServer(script, url)
        self.authenticate(server)

        application_data = server.query(params={"scpv2": "dummy"})
        while True:
            if len(application_data) == 0:
                break
            client_data = self.raw_exchange(application_data)
            application_data = server.query(client_data[:-2], params={"scpv2": "dummy"})
        application_data = server.query(params={"scpv2": "dummy"})

        apps = AppList()
        apps.ParseFromString(application_data)
        for app in apps.list:
            yield AppInfo(app.name, app.flags & 0xFFFF, app.hashCodeData, app.hash)

    @property
    def apps(self):
        data = self.apdu_secure_exchange(LedgerSecureIns.LIST_APPS)
        while len(data) != 0:
            response = ApduListAppsResponse.parse(data)
            for app in response.apps:
                yield AppInfo(
                    app.name, app.flags & 0xFFFF, app.code_data_hash, app.full_hash
                )
            data = self.apdu_secure_exchange(LedgerSecureIns.LIST_APPS_CONTINUE)

    def run_app(self, app_name: str):
        return self.apdu_exchange(LedgerIns.RUN_APP, app_name.encode())



import collections
import colorsys
import gzip
import math
from abc import ABC, abstractmethod
from typing import Dict, List, Optional

from PIL import Image, ImageOps

from ledgerwallet import params
from ledgerwallet.utils import DeviceNames, get_device_name

MAX_COLORS = 16


def is_power2(n):
    return n != 0 and ((n & (n - 1)) == 0)


def _image_to_compressed_buffer_nbgl(im: Image) -> bytes:
    im = im.convert("L")
    nb_colors = len(im.getcolors())

    # Compute bits_per_pixel
    # Round number of colors to a power of 2
    if not is_power2(nb_colors):
        nb_colors = int(pow(2, math.ceil(math.log(nb_colors, 2))))

    bpp = int(math.log(nb_colors, 2))
    # 2 or 3 BPP are not supported
    if bpp > 1:
        bpp = 4

    if bpp == 0:
        bpp = 1

    # Invert if bpp is 1
    if bpp == 1:
        im = ImageOps.invert(im)

    width, height = im.size

    current_byte = 0
    current_bit = 0
    image_data = []
    base_threshold = int(256 / nb_colors)
    half_threshold = int(base_threshold / 2)

    # col first
    for col in reversed(range(width)):
        for row in range(height):
            # Return an index in the indexed colors list
            # top to bottom
            # Perform implicit rotation here (0,0) is left top in NBGL,
            # and generally left bottom for various canvas
            color_index = im.getpixel((col, row))
            color_index = int((color_index + half_threshold) / base_threshold)

            if color_index >= nb_colors:
                color_index = nb_colors - 1

            # le encoded
            current_byte += color_index << ((8 - bpp) - current_bit)
            current_bit += bpp

            if current_bit >= 8:
                image_data.append(current_byte & 0xFF)
                current_bit = 0
                current_byte = 0

    # Handle last byte if any
    if current_bit > 0:
        image_data.append(current_byte & 0xFF)

    # Compress buffer into a gzip file
    output_buffer = []
    # cut into chunks of 2048 bytes max of uncompressed data
    # (because decompression needs the full buffer)
    full_uncompressed_size = len(image_data)
    i = 0
    while full_uncompressed_size > 0:
        chunk_size = min(2048, full_uncompressed_size)
        tmp = bytes(image_data[i : i + chunk_size])
        compressed_buffer = gzip.compress(tmp, mtime=0)
        output_buffer += [
            len(compressed_buffer) & 0xFF,
            (len(compressed_buffer) >> 8) & 0xFF,
        ]
        output_buffer += compressed_buffer
        full_uncompressed_size -= chunk_size
        i += chunk_size

    # Add metadata
    BPP_FORMATS = {1: 0, 2: 1, 4: 2}

    result = [
        width & 0xFF,
        width >> 8,
        height & 0xFF,
        height >> 8,
        (BPP_FORMATS[bpp] << 4)
        | 1,  # 1 is gzip compression type. We only support gzip.
        len(output_buffer) & 0xFF,
        (len(output_buffer) >> 8) & 0xFF,
        (len(output_buffer) >> 16) & 0xFF,
    ]

    result.extend(output_buffer)

    return bytes(bytearray(result))


def _image_to_packed_buffer_bagl(im: Image) -> bytes:
    width, height = im.size
    num_colors = len(im.getcolors())

    # Round number of colors to a power of 2
    if not is_power2(num_colors):
        num_colors = int(pow(2, math.ceil(math.log(num_colors, 2))))

    bits_per_pixel = int(math.log(num_colors, 2))

    current_byte = 0
    current_bit = 0
    image_data = []

    # Reorder color map by luminance
    palette = im.getpalette()
    opalette: Dict[float, List] = {}
    for i in range(num_colors):
        red, green, blue = palette[3 * i : 3 * i + 3]
        hue, saturation, value = colorsys.rgb_to_hsv(
            red / 255.0, green / 255.0, blue / 255.0
        )

        # Several colors could have the same luminance
        if value * 255.0 not in opalette:
            opalette[value * 255.0] = []
        opalette[value * 255.0].append([i, (red << 16) + (green << 8) + blue])
    opalette = collections.OrderedDict(sorted(opalette.items()))

    # Compute the remapping index
    i = 0
    new_indices = {}
    new_palette = []
    for lum, values in opalette.items():
        # Old index to new index
        for v in values:
            new_indices[v[0]] = i
            new_palette.append(v[1])
            i += 1

    palette = new_indices

    # write BPP
    header = bytes([bits_per_pixel])
    # LE color array, it is meant to be embedded as is in an array
    for i in range(num_colors):
        header += new_palette[i].to_bytes(4, "big")

    # Row first
    for row in range(height):
        for col in range(width):
            # Return an index in the indexed colors list for indexed address
            # spaces left to right.
            #
            # Perform implicit rotation here (0,0) is left top in BAGL, and
            # generally left bottom for various canvas.
            color_index = im.getpixel((col, row))

            # Remap index by luminance
            color_index = palette[color_index]

            # le encoded
            current_byte += color_index << current_bit
            current_bit += bits_per_pixel

            if current_bit >= 8:
                image_data.append(current_byte & 0xFF)
                current_bit = 0
                current_byte = 0

    # Handle last byte if any
    if current_bit > 0:
        image_data.append(current_byte & 0xFF)
    return header + bytes(image_data)


def icon_from_file(image_file: str, device: str) -> bytes:
    im = Image.open(image_file)
    im.load()

    assert im.mode == "P" and len(im.getcolors()) <= MAX_COLORS

    if get_device_name(int(device, 16)) in [
        DeviceNames.LEDGER_STAX.value,
        DeviceNames.LEDGER_FLEX.value,
    ]:
        image_data = _image_to_compressed_buffer_nbgl(im)
    else:
        image_data = _image_to_packed_buffer_bagl(im)

    return image_data


class AppManifest(ABC):
    dic: Dict = {}

    @property
    def app_name(self) -> str:
        return self.dic.get("name", "")

    @property
    def target_id(self) -> str:
        return self.dic.get("targetId", "")

    @abstractmethod
    def data_size(self, device: str) -> int:
        pass

    @abstractmethod
    def get_application_flags(self, device: str) -> int:
        pass

    @abstractmethod
    def get_api_level(self, device: str) -> Optional[int]:
        pass

    @abstractmethod
    def get_binary(self, device: str) -> str:
        pass

    @abstractmethod
    def serialize_parameters(self, device: str) -> bytes:
        pass

    @abstractmethod
    def assert_compatible_device(self, device_id: int):
        pass

    def serialize_derivation_path(self, value):
        derivation_paths: Dict[str, Optional[int]] = {
            "paths": None,
            "curve": None,
        }
        for derivation_entry in value:
            if derivation_entry == "curves":
                curves = 0
                for curve in value["curves"]:
                    if curve == "secp256k1":
                        curves |= params.CURVE_SECP256K1
                    elif curve == "prime256r1":
                        curves |= params.CURVE_PRIME256R1
                    elif curve == "ed25519":
                        curves |= params.CURVE_ED25519
                    elif curve == "bls12381g1":
                        curves |= params.CURVE_BLS12381G1
                    derivation_paths["curve"] = curves
            elif derivation_entry == "paths":
                derivation_paths["paths"] = value["paths"]
        return derivation_paths



from ledgerwallet.transport import open_device
from ledgerwallet.transport.device import Device


class MockDevice(Device):
    @classmethod
    def enumerate_devices(cls):
        """Do nothing."""

    def __init__(self):
        self.is_open = False

    def open(self):
        self.is_open = True

    def write(self, data: bytes):
        """Do nothing."""

    def read(self, timeout: int = 0) -> bytes:
        """Do nothing."""

    def exchange(self, data: bytes, timeout: int = 0) -> bytes:
        """Do nothing."""

    def close(self):
        self.is_open = False


def test_context_manager():
    dev = MockDevice()
    assert not dev.is_open

    with open_device(dev):
        assert dev.is_open

    assert not dev.is_open



import os
import sys
from typing import Optional

from ledgerwallet.manifest import AppManifest, icon_from_file
from ledgerwallet.utils import DeviceNames, get_device_name

if sys.version_info >= (3, 11):
    import tomllib
else:
    import toml as tomllib

from ledgerwallet import params


class AppManifestToml(AppManifest):
    def __init__(self, filename):
        self.path = os.path.dirname(filename)
        if sys.version_info >= (3, 11):
            with open(filename, "rb") as f:
                self.dic = tomllib.load(f)
        else:
            with open(filename, "r") as f:
                self.dic = tomllib.load(f)

    def data_size(self, device: str) -> int:
        return self.dic[device].get("dataSize", 0)

    def get_application_flags(self, device: str) -> int:
        return int(self.dic[device].get("flags", "0"), 16)

    def get_api_level(self, device: str) -> Optional[int]:
        if (
            get_device_name(int(device, 16)) == DeviceNames.LEDGER_NANO_SP.value
            and "apiLevel" in self.dic[device]
        ):
            level = self.dic[device]["apiLevel"]
            if isinstance(level, int):
                return int(level)
            return int(level, 10)
        return None

    def get_binary(self, device: str) -> str:
        return os.path.join(self.path, self.dic[device]["binary"])

    def serialize_parameters(self, device: str) -> bytes:
        parameters = []
        for entry, value in self.dic.items():
            if entry == "name":
                parameters.append({"type_": "BOLOS_TAG_APPNAME", "value": value})
            elif entry == "version":
                parameters.append({"type_": "BOLOS_TAG_APPVERSION", "value": value})
            elif entry == device:
                for device_entry, device_value in self.dic[entry].items():
                    if device_entry == "icon":
                        parameters.append(
                            {
                                "type_": "BOLOS_TAG_ICON",
                                "value": icon_from_file(device_value, device),
                            }
                        )
                    elif device_entry == "derivationPath":
                        derivation_paths = self.serialize_derivation_path(device_value)
                        parameters.append(
                            {"type_": "BOLOS_TAG_DERIVEPATH", "value": derivation_paths}
                        )
        return params.AppParams.build(parameters)

    def assert_compatible_device(self, device_id: int):
        if "binary" not in self.dic.get(str(device_id), list()):
            raise ValueError(
                "TOML manifest has no installation information about the current"
                " device : {}".format(get_device_name(device_id))
            )



from unittest import TestCase

from ledgerwallet import utils


class UtilsModuleTest(TestCase):
    def test_serialize(self):
        sample = b"some bytes"
        result = utils.serialize(sample)
        self.assertEqual(result[0], len(result) - 1)
        self.assertEqual(result[0], len(sample))

    def test_unserialize(self):
        sample = bytes.fromhex("0304050607")
        result, rest = utils.unserialize(sample)
        self.assertEqual(result, sample[1 : 1 + sample[0]])
        self.assertEqual(rest, sample[1 + sample[0] :])

    def test_unserialize_too_small(self):
        sample = bytes.fromhex("01")
        with self.assertRaises(AssertionError):
            utils.unserialize(sample)

    def test_flags_to_string(self):
        self.assertEqual(
            utils.flags_to_string(5333),
            "issuer,signed,derive_master,global_pin,debug,custom_ca,no_run",
        )



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Parameter(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    ALIAS_FIELD_NUMBER: builtins.int
    LOCAL_FIELD_NUMBER: builtins.int
    name: typing.Text
    alias: typing.Text
    local: builtins.bool
    def __init__(self,
        *,
        name: typing.Text = ...,
        alias: typing.Text = ...,
        local: builtins.bool = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["alias",b"alias","local",b"local","name",b"name"]) -> None: ...
global___Parameter = Parameter

class Request(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    PARAMETERS_FIELD_NUMBER: builtins.int
    REFERENCE_FIELD_NUMBER: builtins.int
    ELF_FIELD_NUMBER: builtins.int
    CLOSE_FIELD_NUMBER: builtins.int
    LARGESTACK_FIELD_NUMBER: builtins.int
    REMOTE_PARAMETERS_FIELD_NUMBER: builtins.int
    id: typing.Text
    parameters: builtins.bytes
    reference: typing.Text
    elf: builtins.bytes
    close: builtins.bool
    largeStack: builtins.bool
    @property
    def remote_parameters(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Parameter]: ...
    def __init__(self,
        *,
        id: typing.Text = ...,
        parameters: builtins.bytes = ...,
        reference: typing.Text = ...,
        elf: builtins.bytes = ...,
        close: builtins.bool = ...,
        largeStack: builtins.bool = ...,
        remote_parameters: typing.Optional[typing.Iterable[global___Parameter]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["close",b"close","elf",b"elf","id",b"id","largeStack",b"largeStack","parameters",b"parameters","reference",b"reference","remote_parameters",b"remote_parameters"]) -> None: ...
global___Request = Request

class Response(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    RESPONSE_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    EXCEPTION_FIELD_NUMBER: builtins.int
    id: typing.Text
    response: builtins.bytes
    message: typing.Text
    exception: typing.Text
    def __init__(self,
        *,
        id: typing.Text = ...,
        response: builtins.bytes = ...,
        message: typing.Text = ...,
        exception: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["exception",b"exception","id",b"id","message",b"message","response",b"response"]) -> None: ...
global___Response = Response



import configparser
import os
import re
import sys
from json import JSONDecodeError

import click
from tabulate import tabulate

if sys.version_info >= (3, 11):
    from tomllib import TOMLDecodeError
else:
    from toml.decoder import TomlDecodeError as TOMLDecodeError

from ledgerwallet import utils
from ledgerwallet.client import (
    LEDGER_HSM_KEY,
    LEDGER_HSM_URL,
    CommException,
    LedgerClient,
    LedgerIns,
    NoLedgerDeviceException,
)
from ledgerwallet.crypto.ecc import PrivateKey
from ledgerwallet.manifest import AppManifest
from ledgerwallet.manifest_json import AppManifestJson
from ledgerwallet.manifest_toml import AppManifestToml
from ledgerwallet.transport import FileDevice


class ManifestFormatError(Exception):
    def __init__(
        self, toml_error: TOMLDecodeError, json_error: JSONDecodeError
    ) -> None:
        self.toml = toml_error
        self.json = json_error

    def __str__(self) -> str:
        return "TOML or JSON is expected\nTOML error : {}\nJSON error : {}".format(
            self.toml, self.json
        )


_remote_options = [
    click.option("--url", type=str, default=LEDGER_HSM_URL, help="Server URL."),
    click.option(
        "--key",
        "-k",
        type=str,
        default=LEDGER_HSM_KEY,
        help="Key identifier on the remote server.",
    ),
]


def remote_options(func):
    for option in reversed(_remote_options):
        func = option(func)
    return func


def get_private_key() -> bytes:
    app_path = click.get_app_dir("ledgerctl")
    if not os.path.exists(app_path):
        os.makedirs(app_path)

    cfg_file = os.path.join(app_path, "config.ini")
    try:
        config = configparser.RawConfigParser()
        config.read(cfg_file)

        default_config = config["DEFAULT"]
        private_key = bytes.fromhex(default_config["private_key"])
    except KeyError:
        new_private_key = PrivateKey()
        public_key = new_private_key.pubkey
        pubkey_bytes = public_key.serialize(compressed=False)

        config = configparser.RawConfigParser()
        config["DEFAULT"] = {
            "public_key": pubkey_bytes.hex(),
            "private_key": new_private_key.serialize().hex(),
        }
        with click.open_file(cfg_file, "w") as f:
            config.write(f)
        private_key = bytes.fromhex(new_private_key.serialize().hex())

    return private_key


def get_file_device(output_file, target_id="0x33000004"):
    try:
        return LedgerClient(FileDevice(target_id, out=output_file))
    except NoLedgerDeviceException as exception:
        click.echo(exception)
        sys.exit(0)


@click.group()
@click.option("-v", "--verbose", is_flag=True, help="Display exchanged APDU.")
@click.pass_context
def cli(ctx, verbose):
    if verbose:
        utils.enable_apdu_log()

    def get_client():
        try:
            return LedgerClient(private_key=get_private_key())
        except NoLedgerDeviceException as exception:
            click.echo(exception)
            sys.exit(0)

    ctx.obj = get_client


@cli.command(help="Send raw data to the device.")
@click.argument("input_file", type=click.File("r"))
@click.pass_obj
def send(get_client, input_file):
    client = get_client()
    while True:
        chunk = input_file.readline()
        if not chunk:
            break
        response = client.raw_exchange(bytes.fromhex(chunk.rstrip()))
        click.echo(response.hex())


@cli.command(help="Check if device is genuine.")
@remote_options
@click.pass_obj
def genuine_check(get_client, url, key):
    if get_client().genuine_check(url, key):
        click.echo("Device is genuine.")
    else:
        click.echo("Device is NOT genuine.")


@cli.command("list", help="List installed applications.")
@click.option("--remote", default=False, is_flag=True)
@remote_options
@click.pass_obj
def list_apps(get_client, remote, url, key):
    client = get_client()
    rows = []

    # Always list apps using a remote server on Nano X, as custom SCP channels
    # cannot be established
    if client.target_id == 0x33000004:
        remote = True
    for app in client.list_apps_remote(url, key) if remote else client.apps:
        rows.append(
            [
                app.name,
                utils.flags_to_string(app.flags),
                app.code_data_hash.hex(),
                app.full_hash.hex(),
            ]
        )
    if len(rows) == 0:
        click.echo("There is no application on the device.")
    else:
        click.echo(tabulate(rows, ("Name", "Flags", "Code/data hash", "Full hash")))


@cli.command("install", help="Install application.")
@click.argument("manifest")
@click.option(
    "-f",
    "--force",
    help="Delete the app with the same name before loading the provided one.",
    is_flag=True,
)
@click.option(
    "-o",
    "--offline",
    help="Dump APDU installation file, do not attempt to connect to a physical device.",
    is_flag=False,
    flag_value="out.apdu",
)
@click.pass_obj
def install_app(get_client, manifest: AppManifest, force, offline):
    try:
        app_manifest: AppManifest = AppManifestToml(manifest)
    except TOMLDecodeError as toml_error:
        try:
            app_manifest = AppManifestJson(manifest)
            click.echo(
                "[WARNING] JSON files will be deprecated in future version", err=True
            )
        except JSONDecodeError as json_error:
            raise ManifestFormatError(toml_error, json_error)

    try:
        if offline:
            try:
                dump_file = open(offline, "w")
            except OSError:
                click.echo("Unable to open file {} for dump.".format(offline))
                sys.exit(1)
            click.echo("Dumping APDU installation file to {}".format(offline))
            client = get_file_device(dump_file, app_manifest.target_id)
            if force:
                client.delete_app(app_manifest.app_name)
        else:
            client = get_client()
            if force:
                client.delete_app(app_manifest.app_name)
                client.close()
                client = get_client()
        client.install_app(app_manifest)
    except CommException as e:
        if e.sw == 0x6985:
            click.echo("Operation has been canceled by the user.")
        elif e.sw == 0x6A80:
            click.echo("An application with the same name is already installed.")
        elif e.sw == 0x6A81:
            click.echo("Application is already installed.")
        else:
            raise


@cli.command("remote-install", help="Install an application from a remote server.")
@click.argument("app_path")
@click.argument("key_path")
@remote_options
@click.pass_obj
def install_remote_app(get_client, app_path, key_path, url, key):
    client = get_client()
    client.install_remote_app(app_path, key_path, url, key)


@cli.command("delete", help="Delete application.")
@click.argument("app")
@click.option(
    "--by-hash",
    help="Delete using application hash instead of application name",
    is_flag=True,
)
@click.option(
    "-o",
    "--offline",
    help=(
        "Dump APDU delete command file, do not attempt to connect to a physical device."
    ),
    is_flag=False,
    flag_value="out_delete.apdu",
)
@click.pass_obj
def delete_app(get_client, app, by_hash, offline):
    if by_hash:
        data = bytes.fromhex(app)
    else:
        data = app

    if offline:
        try:
            dump_file = open(offline, "w")
        except OSError:
            click.echo("Unable to open file {} for dump.".format(offline))
            sys.exit(1)
        click.echo("Dumping APDU delete command file to {}".format(offline))
        client = get_file_device(dump_file)
    else:
        client = get_client()
    try:
        client.delete_app(data)
    except CommException as e:
        if e.sw == 0x6985:
            click.echo("Operation has been canceled by the user.")
        else:
            raise


@cli.command("remote-delete", help="Delete an application using a remote server.")
@click.argument("app_path")
@click.argument("key_path")
@remote_options
@click.pass_obj
def delete_app_remote(get_client, app_path, key_path, url, key):
    client = get_client()
    client.delete_remote_app(app_path, key_path, url, key)


@cli.command(help="Run specified application.")
@click.argument("app_name")
@click.pass_obj
def run(get_client, app_name: str):
    try:
        get_client().run_app(app_name)
    except CommException as e:
        if e.sw == 0x6984:
            click.echo("Application {} is not installed.".format(app_name))
        else:
            raise


@cli.command(help="Install a custom certificate authority on the device.")
@click.argument("name")
@click.argument("public_key", required=False)
@click.pass_obj
def install_ca(get_client, name, public_key):
    if public_key is None:
        raw_private_key = get_private_key()
        pubkey_bytes = PrivateKey(raw_private_key).pubkey.serialize(compressed=False)
    else:
        pubkey_bytes = bytes.fromhex(public_key)

    try:
        get_client().install_ca(name, pubkey_bytes)
        click.echo("Custom certificate has been successfully installed.")
    except CommException as e:
        if e.sw == 0x6982:
            click.echo("A certificate is already installed on the device.")
        elif e.sw == 0x6802:  # INVALID_PARAMETER
            click.echo("The provided certificate is invalid.")
        elif e.sw == 0x661E:
            click.echo("The device is not in recovery mode.")
        else:
            raise


@cli.command(help="Delete custom certificate authority.")
@click.pass_obj
def delete_ca(get_client):
    try:
        get_client().delete_ca()
        click.echo("Custom certificate has been deleted.")
    except CommException as e:
        if e.sw == 0x6A84:
            click.echo(
                "No custom certificate is installed. There is nothing to delete."
            )
        else:
            raise


# Taken from
# https://web.archive.org/web/20111010015624/http://blogmag.net/blog/read/38/Print_human_readable_file_size
def sizeof_fmt(num, suffix="B"):
    for unit in ("", "K", "M"):
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "G", suffix)


@cli.command(help="Display device memory usage.")
@click.pass_obj
def meminfo(get_client):
    def format_entry(name: str, size: int, max_size: int):
        return "- {0:s}: {1:s} ({2:.2f}%)".format(
            name, sizeof_fmt(size), size / max_size * 100
        )

    memory_info = get_client().get_memory_info()

    total_size = (
        memory_info.system_size + memory_info.applications_size + memory_info.free_size
    )
    click.echo("Memory usage:")
    click.echo(format_entry("System", memory_info.system_size, total_size))
    click.echo(format_entry("Applications", memory_info.applications_size, total_size))
    click.echo(format_entry("Available space", memory_info.free_size, total_size))

    click.echo("")
    click.echo(
        "Installed apps: {} (max: {})".format(
            memory_info.used_app_slots, memory_info.num_app_slots
        )
    )


@cli.command(help="Display device information.")
@click.pass_obj
def info(get_client):
    version_info = get_client().get_version_info()

    click.echo(
        "Device: {} ({})".format(
            utils.get_device_name(version_info.target_id), version_info.target_id
        )
    )
    click.echo("SE version: {}".format(version_info.se_version))
    click.echo("MCU version: {}".format(version_info.mcu_version))

    if version_info.flags["is_onboarded"]:
        click.echo("Device is onboarded.")

    if version_info.flags["recovery_mode"]:
        click.echo("Device is running in RECOVERY mode.")


@cli.command("upgrade-firmware", help="Upgrade firmware.")
@click.argument("firmware_name")
@click.argument("firmware_key")
@remote_options
@click.pass_obj
def upgrade_firmware(get_client, firmware_name, firmware_key, url, key):
    client = get_client()
    client.upgrade_firmware(firmware_name, firmware_key, url, key)


@cli.command(
    help="Perform quick onboarding from command line. Device must be reset and booted in recovery mode (press some buttons during the boot)."  # noqa
)
@click.argument("pin")
@click.argument("words")
@click.option(
    "--id", help="Identity number to initialize", default=0, type=click.IntRange(0, 2)
)
@click.option("--prefix", help="Derivation prefix")
@click.option("--passphrase", help="Derivation passphrase")
@click.pass_obj
def onboard(connect, pin, passphrase, words, id, prefix):
    # Check pin argument
    if not re.match("^[0-9]{4,8}$", pin):
        raise ValueError("Invalid PIN format")

    data = bytearray()
    for string in (pin, prefix, passphrase, words):
        if string:
            as_bytes = string.encode()
            data.append(len(as_bytes))
            data += as_bytes
        else:
            data.append(0)

    client = connect()
    try:
        client.apdu_exchange(LedgerIns.ONBOARD, data, p1=id)
    except CommException as e:
        raise RuntimeError(
            "Onboarding failed (are you in recovery mode?): {:#x}".format(e.sw)
        )


if __name__ == "__main__":
    cli()



import json
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from ledgerwallet.manifest_json import AppManifestJson


class ManifestTestJson(TestCase):
    def setUp(self):
        self.data_dir = Path(__file__).parent / "data"
        self.json_path = self.data_dir / "manifest.json"
        self.assertTrue(self.json_path.is_file())
        self.json_manifest = AppManifestJson(str(self.json_path))

    def test___init__(self):
        self.assertEqual(self.json_manifest.path, str(self.json_path.parent))
        with self.json_path.open() as filee:
            self.assertEqual(self.json_manifest.dic, json.load(filee))

    def test___init__error(self):
        with self.assertRaises(AssertionError):
            AppManifestJson(str(self.data_dir / "empty_manifest.json"))  # missing key

    def test_app_name(self):
        self.assertEqual(self.json_manifest.app_name, "some name")

    def test_data_size(self):
        self.assertEqual(self.json_manifest.data_size(""), 42)

    def test_get_application_flags(self):
        self.assertEqual(self.json_manifest.get_application_flags(""), 0x9999)

    def test_get_binary(self):
        self.assertEqual(
            self.json_manifest.get_binary(""), str(self.data_dir / "some binary")
        )

    def test_get_target_id(self):
        self.json_manifest.assert_compatible_device(0x1234)

    def test_properties_with_minimal_manifest(self):
        manifest_json = AppManifestJson(str(self.data_dir / "minimal_manifest.json"))
        self.assertEqual(manifest_json.app_name, "")
        self.assertEqual(manifest_json.data_size(""), 0)
        self.assertEqual(manifest_json.get_application_flags(""), 0)
        self.assertEqual(
            manifest_json.get_binary(""), str(self.data_dir / "some binary")
        )
        self.assertEqual(manifest_json.serialize_parameters(""), b"")

    def test_serialize_parameters(self):
        # fmt: off
        expected = bytes.fromhex(
            "01" +  # BolosTag 'AppName'
            "09" + "736f6d65206e616d65" +  # "some name"
            "02" +  # BolosTag 'Version'
            "0b" + "322e392e342d6465627567" +  # 2.9.4-debug
            "03" +  # BolosTag 'Icon'
            "04" + "01020304" +  # mocked icon
            "04" +  # BolosTag 'DerivationPath'
            "23" +  # 35: following size
            "05" +  # secp256k1 (1) + ed25519 (4)
            "03" + "8000002c80000000000000ff" +                # "44'/0'/255"
            "05" + "8000002c80000000800000000000000100000190"  # "44'/0'/0'/1/400"
        )
        # fmt: on
        with patch(
            "ledgerwallet.manifest_json.icon_from_file",
            lambda x, y: b"\x01\x02\x03\x04",
        ):
            result_json = self.json_manifest.serialize_parameters("1234")
        self.assertEqual(result_json, expected)



from abc import ABC, abstractmethod


class Device(ABC):
    @classmethod
    @abstractmethod
    def enumerate_devices(cls):
        raise NotImplementedError

    @abstractmethod
    def open(self):
        raise NotImplementedError

    @abstractmethod
    def write(self, data: bytes):
        raise NotImplementedError

    @abstractmethod
    def read(self, timeout: int = 0) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def exchange(self, data: bytes, timeout: int = 0) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: listApps.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0elistApps.proto\x12\x08listapps\"F\n\x03\x41pp\x12\r\n\x05\x66lags\x18\x01 \x01(\r\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\x12\x14\n\x0chashCodeData\x18\x03 \x01(\x0c\x12\x0c\n\x04name\x18\x04 \x01(\t\"&\n\x07\x41ppList\x12\x1b\n\x04list\x18\x01 \x03(\x0b\x32\r.listapps.Appb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'listApps_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _APP._serialized_start = 28
    _APP._serialized_end = 98
    _APPLIST._serialized_start = 100
    _APPLIST._serialized_end = 138
# @@protoc_insertion_point(module_scope)



import json
import os
from typing import Optional

from ledgerwallet import params
from ledgerwallet.manifest import AppManifest, icon_from_file
from ledgerwallet.utils import get_device_name


class AppManifestJson(AppManifest):
    def __init__(self, filename):
        with open(filename) as f:
            self.path = os.path.dirname(filename)
            self.dic = json.load(f)
            assert "targetId" in self.dic and "binary" in self.dic

    def data_size(self, device: str) -> int:
        return int(self.dic.get("dataSize", 0))

    def get_application_flags(self, device: str) -> int:
        return int(self.dic.get("flags", "0"), 16)

    def get_api_level(self, device: str) -> Optional[int]:
        if "apiLevel" not in self.dic:
            return None
        return int(self.dic["apiLevel"], 10)

    def get_binary(self, device: str) -> str:
        return os.path.join(self.path, self.dic["binary"])

    def serialize_parameters(self, device: str) -> bytes:
        parameters = []
        for entry, value in self.dic.items():
            if entry == "name":
                parameters.append({"type_": "BOLOS_TAG_APPNAME", "value": value})
            elif entry == "version":
                parameters.append({"type_": "BOLOS_TAG_APPVERSION", "value": value})
            elif entry == "icon":
                parameters.append(
                    {"type_": "BOLOS_TAG_ICON", "value": icon_from_file(value, device)}
                )
            elif entry == "derivationPath":
                derivation_paths = self.serialize_derivation_path(value)
                parameters.append(
                    {"type_": "BOLOS_TAG_DERIVEPATH", "value": derivation_paths}
                )
        return params.AppParams.build(parameters)

    def assert_compatible_device(self, device_id: int):
        if "targetId" in self.dic and int(self.dic["targetId"], 16) == device_id:
            return
        else:
            raise ValueError(
                "JSON manifest has no installation information about the current"
                " device : {}".format(get_device_name(device_id))
            )



import hid

from .device import Device

LEDGER_VENDOR_ID = 0x2C97


class HidDevice(Device):
    def __init__(self, path):
        self.path = path
        self.device = None
        self.opened = False

    @classmethod
    def enumerate_devices(cls):
        devices = []
        for hidDevice in hid.enumerate(LEDGER_VENDOR_ID, 0):
            if (
                "interface_number" in hidDevice and hidDevice["interface_number"] == 0
            ) or ("usage_page" in hidDevice and hidDevice["usage_page"] == 0xFFA0):
                hid_device_path = hidDevice["path"]
                devices.append(HidDevice(hid_device_path))
        return devices

    def get_name(self):
        return "hid:{}".format(self.path.decode())

    def open(self):
        self.device = hid.device()
        self.device.open_path(self.path)
        self.device.set_nonblocking(True)
        self.opened = True

    def write(self, data: bytes):
        # data is prefixed by its size
        data_to_send = int.to_bytes(len(data), 2, "big") + data
        offset = 0
        seq_idx = 0
        while offset < len(data_to_send):
            # Header: channel (0x101), tag (0x05), sequence index
            header = b"\x01\x01\x05" + seq_idx.to_bytes(2, "big")
            pkt_data = header + data_to_send[offset : offset + 64 - len(header)]

            self.device.write(b"\x00" + pkt_data)
            offset += 64 - len(header)
            seq_idx += 1

    def read(self, timeout: int = 1000) -> bytes:
        seq_idx = 0

        self.device.set_nonblocking(False)
        data_chunk = bytes(self.device.read(64 + 1))
        self.device.set_nonblocking(True)

        assert data_chunk[:2] == b"\x01\x01"
        assert data_chunk[2] == 5
        assert data_chunk[3:5] == seq_idx.to_bytes(2, "big")

        data_len = int.from_bytes(data_chunk[5:7], "big")
        data = data_chunk[7:]

        while len(data) < data_len:
            read_bytes = bytes(self.device.read(64 + 1, timeout_ms=timeout))
            data += read_bytes[5:]

        data = data[:data_len]
        return data

    def exchange(self, data: bytes, timeout: int = 1000):
        self.write(data)
        return self.read(timeout=timeout)

    def close(self):
        if self.opened:
            self.device.close()
        self.opened = False


"""
def getDongle():
    hid_device_path = None
    for hidDevice in hid.enumerate(LEDGER_VENDOR_ID, 0):
        if ('interface_number' in hidDevice and hidDevice['interface_number'] == 0) or (
                'usage_page' in hidDevice and hidDevice['usage_page'] == 0xffa0):
            hid_device_path = hidDevice['path']
    if hid_device_path is not None:
        dev = hid.device()
        dev.open_path(hid_device_path)
        dev.set_nonblocking(True)
        return HidDevice(dev)
    raise Exception("No dongle found")
"""
